#include <icd_module_api.h>

static icd_plugable_fn *icd_module_get_plugable_fns(icd_caller *that);

static int icd_module_command_jade(int fd, int argc, char **argv);

static icd_plugable_fn icd_module_plugable_fns;
static icd_plugable_fn icd_module_agent_plugable_fns;
static icd_plugable_fn icd_module_customer_plugable_fns;

static icd_status init_icd_distributor_jade(icd_distributor *that, char *name, icd_config *data);

static icd_status jade_icd__link_callers_via_pop_customer_find_agent(icd_distributor *dist, void *extra);
static int icd_module__state_bridge_failed(icd_event *event, void *extra); 
static int icd_module__state_channel_failed(icd_event *event, void *extra);
static int icd_module__state_associate_failed(icd_event *event, void *extra);
static int icd_module__limited_ready_state_on_fail(icd_event *event, void *extra);
static int icd_module__pushback_and_ready_on_fail(icd_event *event, void *extra);


static int icd_module__state_ready_start(icd_event *event, void *extra);
static int icd_module__state_ready(icd_event *event, void *extra);
static int icd_module_makecall(int fd, char *id, char *queuename, char *dialstr,char *vars, char *plugablestr);
/* static int icd_module__add_queue(icd_event *event, void *extra); */


int icd_module_load(icd_config_registry *registry) {

	assert(registry != NULL);
	icd_config_registry__register_ptr(registry, "dist", "jade",
									  init_icd_distributor_jade);
	
    icd_command_register("jade",icd_module_command_jade,
     "jade dialer loadable module interface to icd",
    "<cmd> <arg1|arg2|...argN>\n Available cmds:\nmakecall, bridgecalls, unbridgecalls, disconnectcall, getdetails <cmd args>\n eg\n",
    "\nmakecall  \nbridgecalls \nunbridgecalls \ndisconnectcall \ngetdetails ");
    
    
    ast_verbose(VERBOSE_PREFIX_3 "Registered ICD Module[JADE]!\n");

    return 0;
}

int icd_module_unload(void) {
    /* verify you clean up any mem that was created in this module, 
     * exceptions are objects that are part of caller queues or distributotrs they
     *  all have there own clean up
     *               
    */
    icd_plugable__clear_fns(&icd_module_plugable_fns);
    icd_plugable__clear_fns(&icd_module_agent_plugable_fns);
    icd_plugable__clear_fns(&icd_module_customer_plugable_fns);
                
	ast_verbose("External Distributor[JADE] UnLoaded!\n"); 
	return 0;

}

static icd_status init_icd_distributor_jade(icd_distributor *that, char *name, icd_config *data) {
    icd_status result ;
    icd_plugable_fn *plugable_fns;
    char buf[ICD_STRING_LEN];

    assert(that != NULL);
    assert(data != NULL);

    strncpy(that->name,name,sizeof(that->name));
    icd_distributor__set_config_params(that, data);
    icd_distributor__create_lists(that, data);
    /* reverse order  icd_caller__cmp_priority_order, icd_caller__cmp_priority_reverse_order */
    icd_list__set_node_insert_func((icd_list *)that->agents, icd_list__insert_ordered,
          icd_caller__cmp_priority_reverse_order);    
    
    icd_distributor__set_link_callers_fn(that,jade_icd__link_callers_via_pop_customer_find_agent , NULL);
    /* need to set customization points for Distributors plugable functions */
    icd_distributor__set_plugable_fn_ptr(that, &icd_module_get_plugable_fns);
   
    ICD_MEMSET_ZERO(&icd_module_plugable_fns, sizeof(icd_module_plugable_fns));  
    snprintf(buf, sizeof(buf), "JadeDefault");
    result = init_icd_plugable_fns(&icd_module_plugable_fns,buf, data);

    ICD_MEMSET_ZERO(&icd_module_agent_plugable_fns, sizeof(icd_module_agent_plugable_fns));    
    snprintf(buf, sizeof(buf), "JadeAgentDefault");
    result = init_icd_plugable_fns(&icd_module_agent_plugable_fns,buf, data);

    ICD_MEMSET_ZERO(&icd_module_customer_plugable_fns, sizeof(icd_module_customer_plugable_fns));    
    snprintf(buf, sizeof(buf), "JadeCustomerDefault");
    result = init_icd_plugable_fns(&icd_module_customer_plugable_fns,buf, data);
    
    
    /* Now customize plugable interface points for each of the above function ptr strucs */
    plugable_fns = &icd_module_plugable_fns;
    result = icd_plugable__set_state_ready_fn(plugable_fns,icd_module__state_ready_start, NULL);

    plugable_fns = &icd_module_agent_plugable_fns;
    result = icd_plugable__set_state_call_end_fn(plugable_fns, icd_agent__standard_state_call_end, NULL);
    result = icd_plugable__set_cleanup_caller_fn(plugable_fns, icd_agent__standard_cleanup_caller);

    plugable_fns = &icd_module_customer_plugable_fns;
    result = icd_plugable__set_state_call_end_fn(plugable_fns, icd_customer__standard_state_call_end, NULL);
    result = icd_plugable__set_cleanup_caller_fn(plugable_fns, icd_customer__standard_cleanup_caller);
    result = icd_plugable__set_state_ready_fn(plugable_fns, icd_module__state_ready, NULL);
  
    result = icd_plugable__set_state_associate_failed_fn(plugable_fns, icd_module__state_associate_failed, NULL); 
    result = icd_plugable__set_state_channel_failed_fn(plugable_fns, icd_module__state_channel_failed, NULL); 
    result = icd_plugable__set_state_bridge_failed_fn(plugable_fns, icd_module__state_bridge_failed, NULL);
    
    icd_distributor__create_thread(that);

    ast_verbose(VERBOSE_PREFIX_3 "ICD Distributor[%s] Initialized !\n", name);

    
    return ICD_SUCCESS;
}

static icd_plugable_fn *icd_module_get_plugable_fns(icd_caller *that) {
    icd_plugable_fn *plugable_fns = NULL;
    char *dist_name = NULL;
    assert(that != NULL);


    if (icd_caller__get_active_member(that) == NULL) {
        /* always return our standard plugable interface when no distributor */
        plugable_fns = &icd_module_plugable_fns;    

    } else {
        dist_name= vh_read(icd_distributor__get_params(
                           icd_member__get_distributor(icd_caller__get_active_member(that))),"dist");
        /* sanity check did we get called with our distributor dist="agent.priority.group" 
        if  (strcmp(dist_name, ICD_MODULE) == 0) {
        }
        */
        plugable_fns = icd_plugable_fn_list__fetch_fns(icd_caller__get_plugable_fns_list(that), dist_name);    
        if (plugable_fns == NULL) {
            /* if we fetched the caller specific plugab_fnlist for this dist and found 
             * no caller specfic plugable defined 
             * then we just go with our static local interface for agent or customer
            */
            if (icd_caller__has_role(that, ICD_AGENT_ROLE)) 
                plugable_fns = &icd_module_agent_plugable_fns;
            else
                plugable_fns = &icd_module_customer_plugable_fns;
        }

    }
        
    if (plugable_fns == NULL) {
        ast_log(LOG_NOTICE, "Caller %d [%s] has no plugable fn aborting ala crash\n",
           icd_caller__get_id(that), icd_caller__get_name(that));       
    } else {
        ast_log(LOG_NOTICE, "\nCaller %d [%s] state[%s] using icd_module_plugable_fns[%s] ready_fn[%p] for Dist[%s]\n", 
                icd_caller__get_id(that), icd_caller__get_name(that),
                icd_caller__get_state_string(that),
                icd_plugable__get_name(plugable_fns),
                plugable_fns->state_ready_fn, dist_name);
    } 
    assert(plugable_fns != NULL);
    return plugable_fns;

}

/* Match the agents target id with the customers id. */
static icd_status jade_icd__link_callers_via_pop_customer_find_agent(icd_distributor *dist, void *extra) {
    icd_member *customer = NULL;
    icd_member *agent = NULL;
    icd_caller *customer_caller = NULL;
    icd_caller *agent_caller = NULL;
    icd_member_list *agents;
    int cust_id;
    int agent_id;
    int first_agent_id = -1;
    int agent_target_id;
    int cust_target_id;
    int found_match = 0;
    char *tmp_str;
    icd_list_iterator *iter;
    icd_status result;

    assert(dist != NULL);

    /* For each agent on the list, try to find a matching customer. Return
       when first is found and processed so distributor doesn't get hung up. */
    while (found_match == 0) {
        agents = icd_distributor__get_agent_list(dist);
        agent = icd_member_list__pop(agents);
        if (agent != NULL)
            agent_caller = icd_member__get_caller(agent);
        if (agent == NULL || agent_caller == NULL) {
            ast_log(LOG_ERROR, "JADE ICD Distributor %s could not retrieve agent from list\n", icd_distributor__get_name(dist));
            /* What to do with this pseudo-agent? For now, drop on ground. */
            return ICD_ERESOURCE;
        }

        agent_id = icd_caller__get_id(agent_caller);

        /* Make sure we haven't looped through all agents already */
        if (agent_id == first_agent_id) {
            icd_distributor__pushback_agent(dist, agent);
            return ICD_ENOTFOUND;
        }
        if (first_agent_id == -1) {
            first_agent_id = agent_id;
        }

        /* In Autodial Manager we need icd_caller__set_param(agent, "targetid", custid) */
        tmp_str = icd_caller__get_param(agent_caller, "targetid");
        if (tmp_str == NULL) {
            ast_log(LOG_WARNING, "JADE ICD Distributor %s reports that agent %s has no target id\n", icd_distributor__get_name(dist), icd_caller__get_name(agent_caller));
            icd_distributor__pushback_agent(dist, agent);
            return ICD_ERESOURCE;
        }
        agent_target_id = tmp_str ? atoi(tmp_str) : 0;       
        /* Mark it as in use for now, until we see if we have a match */
        result = icd_member__distribute(agent);
        if (result != ICD_SUCCESS) {
            return result;
        }

        /* Go through all the customers looking for a match */
        iter = icd_distributor__get_customer_iterator(dist);
        while (icd_list_iterator__has_more(iter)) {
            customer = (icd_member *)icd_list_iterator__next(iter);
            if (customer != NULL)
                customer_caller = icd_member__get_caller(customer);
            if (customer == NULL || customer_caller == NULL) {
                ast_log(LOG_ERROR, "JADE ICD Distributor %s could not retrieve customer from list\n", icd_distributor__get_name(dist));
                /* Again, drop the pseudo-customer on the ground */
                continue;
            }
            cust_id = icd_caller__get_id(customer_caller);
            if (cust_id == agent_target_id) {
                found_match = 1;
                icd_distributor__remove_caller(dist, customer_caller);
                break;
            }
        }
        destroy_icd_list_iterator(&iter);
        /* If no matching customer, add back to distributor and try the next one. */
        if (found_match == 0) {
            /* Agent list should be FIFO for Autodial so that agent puts self on end. */
            icd_caller__set_state(agent_caller, ICD_CALLER_STATE_READY);
        }
    }

    /* This should not be possible, but in case we bugger up the programming... */
    if (found_match == 0) {
        ast_log(LOG_ERROR, "JADE ICD Distributor %s in impossible state. No recovery attempted.\n", icd_distributor__get_name(dist));
        return ICD_EGENERAL;
    }    

    /* Double check with warning that the customer target ID matches the agent ID */
    tmp_str = icd_caller__get_param(customer_caller, "targetid");
    if (tmp_str == NULL) {
        ast_log(LOG_WARNING, "JADE ICD Distributor %s reports that customer %s has no target id\n", icd_distributor__get_name(dist), icd_caller__get_name(customer_caller));
    }
    cust_target_id = tmp_str ? atoi(tmp_str) : 0;
    if (cust_target_id != agent_id) {
        ast_log(LOG_WARNING, "JADE ICD Distributor %s reports that customer %s target does not match agent %s (though the reverse matches)\n", icd_distributor__get_name(dist), icd_caller__get_name(customer_caller), icd_caller__get_name(agent_caller));
    }

    result = icd_member__distribute(customer);
    if (result != ICD_SUCCESS) {
        icd_caller__set_state(agent_caller, ICD_CALLER_STATE_READY);
        return result;
    }

    result = icd_caller__join_callers(customer_caller, agent_caller);

    /* Figure out who the bridger is, and who the bridgee is */
    result = icd_distributor__select_bridger(agent_caller, customer_caller);

    ast_verbose(VERBOSE_PREFIX_3 "JADE Distributor [%s] Link CustomerID[%d] to AgentID[%d]\n", icd_distributor__get_name(dist),cust_id, agent_id );
    if (icd_caller__has_role(customer_caller, ICD_BRIDGER_ROLE)) {
        result = icd_caller__bridge(customer_caller);
    } else if (icd_caller__has_role(agent_caller, ICD_BRIDGER_ROLE)) {
        result = icd_caller__bridge(agent_caller);
    } else {
        ast_log(LOG_ERROR, "JADE Distributor %s found no bridger responsible to bridge call\n", icd_distributor__get_name(dist));
        return ICD_EGENERAL;
    }

    return ICD_SUCCESS;
}

static int icd_module_command_jade(int fd, int argc, char **argv) {
    static char *help[2] = {"help","jade"};

    if(argc >= 4) {
        if(!strcmp(argv[1],"c") || !strcmp(argv[1],"call") || !strcmp(argv[1],"makecall")) {
          icd_module_makecall(fd,argv[2],argv[3],argv[4],argv[5], argv[6]); 
          
        }
/*        
        if(!strcmp(argv[1],"b") || !strcmp(argv[1],"bridge") || !strcmp(argv[1],"bridgecall")) {
          icd_command_dump_bridgecall(fd, argc, argv);
        }
        if(!strcmp(argv[1],"u") || !strcmp(argv[1],"unbridge") || !strcmp(argv[1],"unbridgecall")) {
          icd_command_dump_unbridgecall(fd, argc, argv);
        }
        if(!strcmp(argv[1],"d") || !strcmp(argv[1],"disconnect") || !strcmp(argv[1],"disconnectcall")) {
          icd_command_dump_disconnectcall(fd, argc, argv);
        }
        if(!strcmp(argv[1],"g") || !strcmp(argv[1],"calldetails") || !strcmp(argv[1],"getcalldetails")) {
          icd_command_dump_calldetails(fd, argc, argv);
        }
*/
        }
    else 
        icd_command_help(fd,2,help);

    /* BCA - What should this return? */
    return ICD_SUCCESS;
    
	return 0;
}
/*
	icd_customer *customer;

	if(argc > 3 && !strcasecmp(argv[1],"makecall")) {
		customer = icd_customer__generate_queued_call(argv[2],argv[3],argv[4],icd_module_func_finder_cust);
		if(!customer) {
			ast_cli(fd,"Call Failed\n");
			return -1;
		}
*/	
static int icd_module_makecall(int fd, char *id, char *queuename, char *dialstr,char *vars, char *plugablestr) {

	return 0;
}

static int icd_module__state_bridge_failed(icd_event *event, void *extra) {
    icd_caller *that=NULL;

    assert(event != NULL);

    /* Actions for when the bridge failed:
     *   - set state to READY (and thus back into distributor)
     *   - set state to READY if bridge failed < 5 times
     *   - set pushed_back attribute and then set state to READY
     *   - destroy caller
     *
     */
    that = (icd_caller *)icd_event__get_source(event);
    assert(that != NULL);
    icd_caller__set_state_on_associations(that,ICD_CALLER_STATE_BRIDGE_FAILED);

    return icd_module__pushback_and_ready_on_fail(event, extra);

}

static int icd_module__state_channel_failed(icd_event *event, void *extra) {
    assert(event != NULL);

    return icd_module__limited_ready_state_on_fail(event, extra);
}

static int icd_module__state_associate_failed(icd_event *event, void *extra) {
    assert(event != NULL);
    /*
     that = (icd_caller *)icd_event__get_source(event);
     assert(that != NULL);
    */

    return icd_module__pushback_and_ready_on_fail(event, extra);
}

static int icd_module__pushback_and_ready_on_fail(icd_event *event, void *extra) {
    icd_caller *that;
    int *value = 0;

    assert(event != NULL);

    that = icd_event__get_source(event);
    assert(that != NULL);

    /* increment the count of the number of failures */ 
    value = (int *)extra;
    
    //(*value)++;

    icd_caller__set_pushback(that);
    icd_caller__set_state(that, ICD_CALLER_STATE_READY);

    /* No veto allowed here */
    return 0;
}


static int icd_module__limited_ready_state_on_fail(icd_event *event, void *extra) {
    icd_caller *that;
    int *value;

    assert(event != NULL);
    assert(extra != NULL);

    that = icd_event__get_source(event);
    assert(that != NULL);

    value = (int *)extra;

    /*
    TBD - Set up cleanup thread and add to list it keeps It will call icd_caller__clear()
    (*value)++;
    if ((*value) < 5) {
        icd_caller__set_state(that, ICD_CALLER_STATE_READY);
    } else {
    }
    on channel state failed, as of now,
    simply call the standard cleanup caller which will put the agent back to the q
    and customer is hung up with his channel torn down.
    More complex cases are yet to be dealt with
    */
    icd_caller__set_pushback(that);
    /* that->cleanup_caller_fn(that); standard is in icd_agent/customer.c */
    /* No veto allowed here */
    return 0;
}

static int icd_module__state_ready_start(icd_event *event, void *extra) {
    icd_caller *caller;
	static int waitms = 100;
	int res=0,result=0;
	ast_channel *chan;

	assert(event != NULL);
    caller = (icd_caller *)icd_event__get_source(event);
	chan = icd_caller__get_channel(caller);
	result = icd_caller__dial_channel(caller);
	if(!chan)
		return 0;

    /* clear all association b4 push onto Distributor since dist will set associations */
    icd_caller__remove_all_associations(caller);

	for(;;) {
		for (;;) {
			if (!(icd_caller__get_state(caller) == ICD_CALLER_STATE_READY)) {
				/*state change occured, we are just a bridgee waiting to get bridged continue waiting */
				if (!(icd_caller__has_role(caller, ICD_BRIDGEE_ROLE) && 
					  (icd_caller__get_state(caller) == ICD_CALLER_STATE_GET_CHANNELS_AND_BRIDGE ||
					   icd_caller__get_state(caller) == ICD_CALLER_STATE_DISTRIBUTING)) 
					) {
					result = -1;
					res = 0;
					break;
				}
			}
			/* Wait some milliseconds before checking again */
			res = ast_waitfordigit(chan, waitms);
			/* -1 hangup detected, or DTMF digits #-35,*-42 0-48,1-49, ..9-57 or t-116 timeout */
			if(res != 0) 
				break;
		}

		if (res < 0) {
			icd_bridge__safe_hangup(caller);
			icd_caller__set_state(caller, ICD_CALLER_STATE_CALL_END);
			break;
		}
		
		if (result < 0) {
			/* state change occured, time to go to next state */
			break;
		}
		

	}

    return 0;

}
    
/* Standard function for reacting to the ready state */
/*!
 * Required state in order to be available for distributor to link. If not in
 * this state, distributor discards member when popped.
 */
static int icd_module__state_ready(icd_event *event, void *extra) {
    icd_caller *caller;
	static int waitms = 100;
	int res=0,result=0;
	ast_channel *chan;

	assert(event != NULL);
    caller = (icd_caller *)icd_event__get_source(event);
	chan = icd_caller__get_channel(caller);
	result = icd_caller__dial_channel(caller);
	if(!chan)
		return 0;

    /* clear all association b4 push onto Distributor since dist will set associations */
    icd_caller__remove_all_associations(caller);

	for(;;) {
		for (;;) {
			if (!(icd_caller__get_state(caller) == ICD_CALLER_STATE_READY)) {
				/*state change occured, we are just a bridgee waiting to get bridged continue waiting */
				if (!(icd_caller__has_role(caller, ICD_BRIDGEE_ROLE) && 
					  (icd_caller__get_state(caller) == ICD_CALLER_STATE_GET_CHANNELS_AND_BRIDGE ||
					   icd_caller__get_state(caller) == ICD_CALLER_STATE_DISTRIBUTING)) 
					) {
					result = -1;
					res = 0;
					break;
				}
			}
			/* Wait some milliseconds before checking again */
			res = ast_waitfordigit(chan, waitms);
			/* -1 hangup detected, or DTMF digits #-35,*-42 0-48,1-49, ..9-57 or t-116 timeout */
			if(res != 0) 
				break;
		}

		if (res < 0) {
			icd_bridge__safe_hangup(caller);
			icd_caller__set_state(caller, ICD_CALLER_STATE_CALL_END);
			break;
		}
		
		if (result < 0) {
			/* state change occured, time to go to next state */
			break;
		}
		

	}

    return 0;
}



